@@ -19,21 +19,25 @@
 # License along with this program; if not, write to the
 # Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 # Boston, MA 02110-1301, USA.
-#
-# SPDX-License-Identifier: LGPL-2.0-or-later
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 3, or (at your option)
+# any later version.
 
 import sys
 import inspect
-import itertools
-import weakref
 from ..overrides import override
-from ..module import get_introspection_module
-
-from gi.repository import GLib
+from ..importer import modules
 
 
-Gst = get_introspection_module('Gst')
+if sys.version_info >= (3, 0):
+    _basestring = str
+    _callable = lambda c: hasattr(c, '__call__')
+else:
+    _basestring = basestring
+    _callable = callable
 
+Gst = modules['Gst']._introspection_module
 __all__ = []
 
 if Gst._version == '0.10':
@@ -42,27 +46,11 @@ if Gst._version == '0.10':
 was not designed for use with introspection some of the \
 interfaces and API will fail.  As such this is not supported \
 by the GStreamer development team and we encourage you to \
-port your app to Gst 1 or greater. gst-python is the recommended \
+port your app to Gst 1 or greater. gst-python is the recomended \
 python module to use with Gst 0.10"
 
     warnings.warn(warn_msg, RuntimeWarning)
 
-
-class Element(Gst.Element):
-    @staticmethod
-    def link_many(*args):
-        '''
-        @raises: Gst.LinkError
-        '''
-        for pair in pairwise(args):
-            if not pair[0].link(pair[1]):
-                raise LinkError(
-                    'Failed to link {} and {}'.format(pair[0], pair[1]))
-
-Element = override(Element)
-__all__.append('Element')
-
-
 class Bin(Gst.Bin):
     def __init__(self, name=None):
         Gst.Bin.__init__(self, name=name)
@@ -72,18 +60,6 @@ class Bin(Gst.Bin):
             if not Gst.Bin.add(self, arg):
                 raise AddError(arg)
 
-    def make_and_add(self, factory_name, instance_name=None):
-        '''
-        @raises: Gst.AddError
-        '''
-        elem = Gst.ElementFactory.make(factory_name, instance_name)
-        if not elem:
-            raise AddError(
-                'No such element: {}'.format(factory_name))
-        self.add(elem)
-        return elem
-
-
 Bin = override(Bin)
 __all__.append('Bin')
 
@@ -92,24 +68,15 @@ class Caps(Gst.Caps):
     def __nonzero__(self):
         return not self.is_empty()
 
-    def __new__(cls, *args):
-        if not args:
+    def __new__(cls, *kwargs):
+        if not kwargs:
             return Caps.new_empty()
-        elif len(args) > 1:
+        elif len(kwargs) > 1:
             raise TypeError("wrong arguments when creating GstCaps object")
-        elif isinstance(args[0], str):
-            return Caps.from_string(args[0])
-        elif isinstance(args[0], Caps):
-            return args[0].copy()
-        elif isinstance(args[0], Structure):
-            res = Caps.new_empty()
-            res.append_structure(args[0])
-            return res
-        elif isinstance(args[0], (list, tuple)):
-            res = Caps.new_empty()
-            for e in args[0]:
-                res.append_structure(e)
-            return res
+        elif isinstance(kwargs[0], str):
+            return Caps.from_string(kwargs[0])
+        elif isinstance(kwargs[0], Caps):
+            return kwargs[0].copy()
 
         raise TypeError("wrong arguments when creating GstCaps object")
 
@@ -130,65 +97,59 @@ class Caps(Gst.Caps):
 Caps = override(Caps)
 __all__.append('Caps')
 
-class PadFunc:
-    def __init__(self, func):
-        self.func = func
+class Pad(Gst.Pad):
+    def __init__(self, *args, **kwargs):
+        self._real_chain_func = None
+        self._real_event_func = None
+        self._real_query_func = None
+        super(Gst.Pad, self).__init__(*args, **kwargs)
 
-    def __call__(self, pad, parent, obj):
-        if isinstance(self.func, weakref.WeakMethod):
-            func = self.func()
-        else:
-            func = self.func
+    def _chain_override(self, pad, parent, buf):
+        return self._real_chain_func(pad, buf)
 
+    def _event_override(self, pad, parent, event):
+        return self._real_event_func(pad, event)
+
+    def _query_override(self, pad, parent, query):
+        query.mini_object.refcount -= 1
         try:
-            res = func(pad, obj)
+            res = self._real_query_func(pad, query)
         except TypeError:
             try:
-                res = func(pad, parent, obj)
+                res = self._real_query_func(pad, parent, query)
             except TypeError:
-                raise TypeError("Invalid method %s, 2 or 3 arguments required"
-                                % func)
+                raise TypeError("Invalid query method %s, 2 or 3 arguments required"
+                                % self._real_query_func)
+        query.mini_object.refcount += 1
 
         return res
 
-class Pad(Gst.Pad):
-    def __init__(self, *args, **kwargs):
-        super(Gst.Pad, self).__init__(*args, **kwargs)
-
     def set_chain_function(self, func):
-        self.set_chain_function_full(PadFunc(func), None)
+        self._real_chain_func = func
+        self.set_chain_function_full(self._chain_override, None)
 
     def set_event_function(self, func):
-        self.set_event_function_full(PadFunc(func), None)
+        self._real_event_func = func
+        self.set_event_function_full(self._event_override, None)
 
     def set_query_function(self, func):
-        self.set_query_function_full(PadFunc(func), None)
+        self._real_query_func = func
+        self.set_query_function_full(self._chain_override, None)
+
+    def set_query_function_full(self, func, udata):
+        self._real_query_func = func
+        self._real_set_query_function_full(self._query_override, None)
 
     def query_caps(self, filter=None):
         return Gst.Pad.query_caps(self, filter)
 
-    def set_caps(self, caps):
-        if not isinstance(caps, Gst.Caps):
-            raise TypeError("%s is not a Gst.Caps." % (type(caps)))
-
-        if not caps.is_fixed():
-            return False
-
-        event = Gst.Event.new_caps(caps)
-
-        if self.direction == Gst.PadDirection.SRC:
-            res = self.push_event(event)
-        else:
-            res = self.send_event(event)
-
-        return res
-
     def link(self, pad):
         ret = Gst.Pad.link(self, pad)
         if ret != Gst.PadLinkReturn.OK:
             raise LinkError(ret)
         return ret
 
+Pad._real_set_query_function_full = Gst.Pad.set_query_function_full
 Pad = override(Pad)
 __all__.append('Pad')
 
@@ -223,11 +184,6 @@ class LinkError(Exception):
     pass
 __all__.append('LinkError')
 
-class MapError(Exception):
-    pass
-__all__.append('MapError')
-
-
 class Iterator(Gst.Iterator):
     def __iter__(self):
         while True:
@@ -268,48 +224,12 @@ Pipeline = override(Pipeline)
 __all__.append('Pipeline')
 
 class Structure(Gst.Structure):
-    def __new__(cls, *args, **kwargs):
-        if not args:
-            if kwargs:
-                raise TypeError("wrong arguments when creating GstStructure, first argument"
-                                " must be the structure name.")
-            return Structure.new_empty()
-        elif len(args) > 1:
-            raise TypeError("wrong arguments when creating GstStructure object")
-        elif isinstance(args[0], str):
-            if not kwargs:
-                return Structure.from_string(args[0])[0]
-            struct = Structure.new_empty(args[0])
-            for k, v in kwargs.items():
-                struct[k] = v
-
-            return struct
-        elif isinstance(args[0], Structure):
-            return args[0].copy()
-
-        raise TypeError("wrong arguments when creating GstStructure object")
-
-    def __init__(self, *args, **kwargs):
-        pass
-
     def __getitem__(self, key):
         return self.get_value(key)
 
-    def keys(self):
-        keys = set()
-        def foreach(fid, value, unused1, udata):
-            keys.add(GLib.quark_to_string(fid))
-            return True
-
-        self.foreach(foreach, None, None)
-        return keys
-
     def __setitem__(self, key, value):
         return self.set_value(key, value)
 
-    def __str__(self):
-        return self.to_string()
-
 Structure = override(Structure)
 __all__.append('Structure')
 
@@ -428,16 +348,8 @@ class IntRange(Gst.IntRange):
             return '[%d,%d,%d]' % (self.range.start, self.range.stop,
                     self.range.step)
 
-    def __eq__(self, other):
-        if isinstance(other, range):
-            return self.range == other
-        elif isinstance(other, IntRange):
-            return self.range == other.range
-        return False
-
-if sys.version_info >= (3, 0):
-    IntRange = override(IntRange)
-    __all__.append('IntRange')
+IntRange = override(IntRange)
+__all__.append('IntRange')
 
 
 class Int64Range(Gst.Int64Range):
@@ -467,34 +379,9 @@ class Int64Range(Gst.Int64Range):
             return '(int64)[%d,%d,%d]' % (self.range.start, self.range.stop,
                     self.range.step)
 
-    def __eq__(self, other):
-        if isinstance(other, range):
-            return self.range == other
-        elif isinstance(other, IntRange):
-            return self.range == other.range
-        return False
 
-class Bitmask(Gst.Bitmask):
-    def __init__(self, v):
-        if not isinstance(v, int):
-            raise TypeError("%s is not an int." % (type(v)))
-
-        self.v = int(v)
-
-    def __str__(self):
-        return hex(self.v)
-
-    def __eq__(self, other):
-        return self.v == other
-
-
-Bitmask = override(Bitmask)
-__all__.append('Bitmask')
-
-
-if sys.version_info >= (3, 0):
-    Int64Range = override(Int64Range)
-    __all__.append('Int64Range')
+Int64Range = override(Int64Range)
+__all__.append('Int64Range')
 
 
 class DoubleRange(Gst.DoubleRange):
@@ -586,80 +473,6 @@ class ValueList(Gst.ValueList):
 ValueList = override(ValueList)
 __all__.append('ValueList')
 
-# From https://docs.python.org/3/library/itertools.html
-
-
-def pairwise(iterable):
-    a, b = itertools.tee(iterable)
-    next(b, None)
-    return zip(a, b)
-
-class MapInfo:
-    def __init__(self):
-        self.memory = None
-        self.flags = Gst.MapFlags(0)
-        self.size = 0
-        self.maxsize = 0
-        self.data = None
-        self.user_data = None
-        self.__parent__ = None
-
-    def __iter__(self):
-        # Make it behave like a tuple similar to the PyGObject generated API for
-        # the `Gst.Buffer.map()` and friends.
-        for i in (self.__parent__ is not None, self):
-            yield i
-
-    def __enter__(self):
-        if not self.__parent__:
-            raise MapError('MappingError', 'Mapping was not successful')
-
-        return self
-
-    def __exit__(self, type, value, tb):
-        if not self.__parent__.unmap(self):
-            raise MapError('MappingError', 'Unmapping was not successful')
-
-__all__.append("MapInfo")
-
-class Buffer(Gst.Buffer):
-
-    def map_range(self, idx, length, flags):
-        mapinfo = MapInfo()
-        if (_gi_gst.buffer_override_map_range(self, mapinfo, idx, length, int(flags))):
-            mapinfo.__parent__ = self
-
-        return mapinfo
-
-    def map(self, flags):
-        mapinfo = MapInfo()
-        if _gi_gst.buffer_override_map(self, mapinfo, int(flags)):
-            mapinfo.__parent__ = self
-
-        return mapinfo
-
-    def unmap(self, mapinfo):
-        mapinfo.__parent__ = None
-        return _gi_gst.buffer_override_unmap(self, mapinfo)
-
-Buffer = override(Buffer)
-__all__.append('Buffer')
-
-class Memory(Gst.Memory):
-
-    def map(self, flags):
-        mapinfo = MapInfo()
-        if (_gi_gst.memory_override_map(self, mapinfo, int(flags))):
-            mapinfo.__parent__ = self
-
-        return mapinfo
-
-    def unmap(self, mapinfo):
-        mapinfo.__parent__ = None
-        return _gi_gst.memory_override_unmap(self, mapinfo)
-
-Memory = override(Memory)
-__all__.append('Memory')
 
 def TIME_ARGS(time):
     if time == Gst.CLOCK_TIME_NONE:
@@ -671,7 +484,7 @@ def TIME_ARGS(time):
                                   time % Gst.SECOND)
 __all__.append('TIME_ARGS')
 
-from gi.overrides import _gi_gst
+from . import _gi_gst
 _gi_gst
 
 # maybe more python and less C some day if core turns a bit more introspection
@@ -715,7 +528,7 @@ def init_pygst():
 
 def deinit_pygst():
     for fname, func in real_functions:
-        if fname not in ["init", "init_check", "deinit", "is_initialized"]:
+        if fname not in ["init", "init_check", "deinit"]:
             setattr(Gst, fname, fake_method)
     for cname_class, methods in class_methods:
         for mname, method in methods:
@@ -724,22 +537,13 @@ def deinit_pygst():
 real_init = Gst.init
 def init(argv):
     init_pygst()
-
-    if Gst.is_initialized():
-        return True
-
     return real_init(argv)
-
 Gst.init = init
 
 real_init_check = Gst.init_check
 def init_check(argv):
     init_pygst()
-    if Gst.is_initialized():
-        return True
-
     return real_init_check(argv)
-
 Gst.init_check = init_check
 
 real_deinit = Gst.deinit
@@ -747,14 +551,6 @@ def deinit():
     deinit_pygst()
     return real_deinit()
 
-def init_python():
-    if not Gst.is_initialized():
-        raise NotInitialized("Gst.init_python should never be called before GStreamer itself is initialized")
-
-    init_pygst()
-
 Gst.deinit = deinit
-Gst.init_python = init_python
 
-if not Gst.is_initialized():
-    deinit_pygst()
+deinit_pygst()
